import collections.abc
import getpass

import pyrfc

from .exception import LogonError, CommunicationError
from ._function import Function
from ._landscape import Landscape
from ._table import select
from netlink.logging import logger

default_landscape = Landscape()


class Server(collections.abc.Mapping):
    PARAMETER_NAMES = ('ashost', 'sysnr', 'mshost', 'msserv', 'sysid', 'group')

    def __init__(self,
                 ashost: str = None, sysnr: str = None,  # direct Application Host
                 mshost: str = None, msserv: str = None, sysid: str = None, group: str = None,  # via message server
                 ):
        """

        :param ashost:
        :param sysnr:
        :param mshost:
        :param msserv:
        :param sysid:
        :param group:
        """
        self.ashost = ashost
        self.sysnr = sysnr
        self.mshost = mshost
        self.msserv = msserv
        self.sysid = sysid
        self.group = group

    def __getitem__(self, item):
        if item.lower() in self.PARAMETER_NAMES:
            return self.__dict__[item.lower()]
        raise KeyError(item)

    def __getattr__(self, item):
        if item.lower() in self.PARAMETER_NAMES:
            return self[item]
        raise AttributeError(item)

    def __len__(self):
        return len([i for i in self.PARAMETER_NAMES if self[i]])

    def __iter__(self):
        return iter([i for i in self.PARAMETER_NAMES if self[i]])

    def kwargs(self):
        return {k: v for k, v in self.items()}

    def __str__(self):
        if self.ashost is not None:
            return f'{self.ashost}:32{self.sysnr}'
        else:
            return f"{self.mshost}:{self.msserv} ({self.sysid}-{self.group})"

    @classmethod
    def from_landscape(cls,
                       sysid: str,
                       landscape: Landscape = None) -> 'Server':
        """
        Get server info from Landscape xml

        :param sysid:
        :param landscape:
        """
        if landscape is None:
            landscape = default_landscape
        try:
            entry = landscape[sysid.upper()]
        except KeyError:
            msg = f"System ID '{sysid.upper()}' not found in Landscape."
            logger.error(msg)
            raise KeyError(msg) from None
        kwargs = {k: v for k, v in entry.items() if k in cls.PARAMETER_NAMES}
        return cls(**kwargs)


class Connection:
    def __init__(self,
                 server: Server,
                 client: str,
                 user: str = None,
                 passwd: str = None,
                 snc_qop: str = None,
                 snc_myname: str = None,
                 snc_partnername: str = None,
                 language: str = 'EN',
                 raw: bool = False):
        kwargs = server.kwargs()
        kwargs['client'] = client
        # fmt: off
        if user is not None: kwargs['user'] = user
        if passwd is not None: kwargs['passwd'] = passwd
        if snc_qop is not None: kwargs['snc_qop'] = snc_qop
        if snc_myname is not None: kwargs['snc_myname'] = snc_myname
        if snc_partnername is not None: kwargs['snc_partnername'] = snc_partnername
        # fmt: on
        kwargs.update(dict(language=language, config=dict(dtime=not raw)))
        logger.debug(f'Connecting to {server}')
        if logger.level < 10:
            for k, v in kwargs.items():
                # fmt: off
                if k == 'passwd': v = '*' * 8
                # fmt: on
                logger.trace(f'{k}: {v}')
        try:
            self._connection = pyrfc.Connection(**kwargs)
        except pyrfc.LogonError as e:
            logger.error(e.message)
            raise LogonError from None
        except pyrfc.CommunicationError as e:
            logger.error(e.message)
            raise CommunicationError from None

        self.connection_attributes = self._connection.get_connection_attributes()
        self._functions = {}

    def __getattr__(self, item):
        if item == "call":
            return self._connection.call
        if item == "get_function_description":
            return self._connection.get_function_description
        if item.upper() not in self._functions and item in self.connection_attributes:
            return self.connection_attributes[item]
        return self[item]

    def __getitem__(self, item):
        item = item.upper()
        if item not in self._functions:
            logger.debug(f"Initializing function '{item}'")
            self._functions[item] = Function(self, item)
        return self._functions[item]

    @property
    def sid(self):
        return self.sysId

    @property
    def sysid(self):
        return self.sysId

    def __str__(self):
        return f"{self.sysId}/{self.client} ({self.user})"

    def close(self):
        self._connection.close()

    def __del__(self):
        self.close()

    def select(self, table, *args, **kwargs):
        return select(self, table, *args, **kwargs)

    # def xml_select(self, table, where):
    #     return xml_select(self, table, where)


def sso(sysid: str, client: str, user: str = None, language="EN", raw: bool = False):
    """
    Connect to SAP using Single-Sign-On

    :param sysid: System ID (<sid>)
    :param client: SAP Client
    :param user: User ID, defaults currently user
    :param language: Default: EN
    :param raw: Default: False
    :return: sap.rfc.Connection
    """
    sysid = sysid.upper()
    if user is None:
        user = getpass.getuser()
    user = user.upper()

    login_info = default_landscape[sysid].copy()
    if not login_info.get("sncname"):
        msg = f"SNC Name for {sysid} not found."
        logger.error(msg)
        raise AttributeError(msg)

    logger.verbose(f"Connecting using SSO to {sysid}/{client} with {user}")
    return Connection(server=Server.from_landscape(sysid),
                      client=client,
                      user=user,
                      snc_qop='9',
                      snc_partnername=login_info["sncname"],
                      language=language,
                      raw=raw)


def login(server: Server, client: str, passwd: str, user: str = None, language="EN", raw: bool = False):
    """
    Connect to SAP

    :param server:
    :param client:
    :param passwd:
    :param user:
    :param language:
    :param raw:
    :return:
    """
    if user is None:
        user = getpass.getuser()
    user = user.upper()

    logger.verbose(f"Connecting to {server} with {user}")
    return Connection(server=server,
                      client=client,
                      user=user,
                      passwd=passwd,
                      language=language,
                      raw=raw)


def login_sid(sysid: str, client: str, passwd: str, user: str = None, language="EN", raw: bool = False):
    """

    :param sysid:
    :param client:
    :param passwd:
    :param user:
    :param language:
    :param raw:
    :return:
    """
    sysid = sysid.upper()
    return login(server=Server.from_landscape(sysid),
                 client=client,
                 user=user,
                 passwd=passwd,
                 language=language,
                 raw=raw)



import collections.abc


def is_dict(value):
    return callable(getattr(value, "keys", False))


def is_list(value):
    if isinstance(value, (str, bytes)) or is_dict(value):
        return False
    return callable(getattr(value, "__iter__", False))


def normalize_dict(value):
    result = {}
    for i in value:
        if is_dict(value[i]):
            result[i.upper()] = normalize_dict(value[i])
        elif is_list(value[i]):
            result[i.upper()] = [normalize_dict(j) for j in value[i]]
        else:
            result[i.upper()] = value[i]
    return result


class ResultDict(collections.abc.MutableMapping):
    def __init__(self, result):
        self._data = {}
        for i in result:
            if is_dict(result[i]):
                self._data[i] = ResultDict(result[i])
            elif is_list(result[i]):
                self._data[i] = [ResultDict(j) for j in result[i]]
            else:
                self._data[i] = result[i]

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        return self._data[item.upper()]

    def __setitem__(self, item, value):
        self._data[item.upper()] = value

    def __delitem__(self, item):
        del self._data[item.upper()]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)


class Function:
    def __init__(self, connection, function):
        self._connection = connection
        self._name = function.upper()
        self._description = None

    # @property
    # def description(self):
    #     if self._description is None:
    #         self._description = Description(connection=self._connection, function_name=self._name)
    #     return self._description

    def __call__(self, **kwargs):
        call_kwargs = normalize_dict(kwargs)
        result = self._connection.call(self._name, **call_kwargs)
        return ResultDict(result)



import collections.abc
import re
import textwrap

from netlink.logging import logger


def select(rfc_connection, table, *args, **kwargs):
    arguments = {'QUERY_TABLE': table.upper()}
    if args:
        arguments['FIELDS'] = [{'FIELDNAME': i.upper()} for i in args]
    if kwargs:
        options_items = []
        for i in kwargs:
            if i.upper() == '_SKIP':
                arguments['ROWSKIPS'] = kwargs[i]
                continue
            if i.upper() == '_COUNT':
                arguments['ROWCOUNT'] = kwargs[i]
                continue
            if isinstance(kwargs[i], int):
                options_items.append(f"{i.upper()} EQ {kwargs[i]}")
            else:
                if '*' in kwargs[i]:
                    options_items.append(f"{i.upper()} LIKE '{kwargs[i].replace('*', '%')}'")
                else:
                    options_items.append(f"{i.upper()} EQ '{kwargs[i]}'")
        arguments['OPTIONS'] = [{'TEXT': i} for i in
                                textwrap.wrap(' AND '.join(options_items))]
    logger.debug(f'{arguments}')
    rfc_response = rfc_connection.rfc_read_table(**arguments)
    regex = re.compile(''.join(['(.{%s})' % int(i.LENGTH) for i in rfc_response.FIELDS]))
    length = sum([int(i.LENGTH) for i in rfc_response.FIELDS])
    fields = tuple([i.FIELDNAME for i in rfc_response.FIELDS])
    result = []
    for data in rfc_response.DATA:
        line = data.WA + ' '*(length-len(data.WA))
        m = regex.match(line)
        if m:
            rcd = {}
            for i in range(m.lastindex):
                rcd[rfc_response.FIELDS[i].FIELDNAME] = m.group(i+1).rstrip()
            result.append(Record(fields, **rcd))
    return result


class Record(collections.abc.Mapping):
    def __init__(self, fields, **kwargs):
        self._fields = fields
        self._data = {}
        for kw in kwargs:
            self._data[kw.upper()] = kwargs[kw]

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        if isinstance(item, int):
            # get by index in field-list
            item = self._fields[item]
        return self._data[item.upper()]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)



"""Get system information from the SAPGUI landscape configuration

For the time being, this works on Windows only, and expects that SAPGUI is installed and configured.
"""

import collections.abc
import winreg

from lxml import etree

from netlink.logging import logger


def _get_landscape_files():
    registry_key = winreg.OpenKey(winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER),
                                  r"Software\SAP\SAPLogon\LandscapeFilesLastUsed")
    result = []
    try:
        path = winreg.QueryValueEx(registry_key, "LandscapeFileOnServer")[0]
        logger.trace(f'Found SAPGUI Landscape Server File {path}')
        result.append(path)
    except FileNotFoundError:
        logger.verbose('No entry for SAPgui Landscape Server File found.')
    try:
        path = winreg.QueryValueEx(registry_key, "LandscapeFile")[0]
        logger.trace(f'Found local SAPGUI Landscape File {path}')
        result.append(path)
    except FileNotFoundError:
        logger.verbose('No entry for local SAPgui Landscape File found.')
    return tuple(result)


def _process_landscape_xml(file_name):
    with open(file_name, "r", encoding="utf-8-sig") as f:
        root = etree.parse(f).getroot()
    if root.tag != "Landscape":
        msg = f"Wrong root tag: <{root.tag}> - expected <Landscape>"
        logger.error(msg)
        raise ValueError(msg)
    sysids = collections.defaultdict(dict)
    message_servers = {}
    for element in root:
        if element.tag == "Services":
            for child in element:
                # noinspection PyProtectedMember
                if type(child) == etree._Comment:
                    continue
                if child.tag != "Service":
                    logger.warning(f"Wrong tag: <{child.tag}> - expected <Service>")
                    continue
                if child.get("type") == "SAPGUI":
                    sid = child.get("systemid")
                    if child.get("sncname"):
                        sysids[sid]["sncname"] = child.get("sncname")
                    if child.get("mode") == "1":
                        sysids[sid]["ashost"], sysids[sid]["sysnr"] = child.get("server").split(":", 1)
                        sysids[sid]["sysnr"] = sysids[sid]["sysnr"][-2:]
                    else:
                        sysids[sid]["msid"] = child.get("msid")
                        sysids[sid]["group"] = child.get("server")
        elif element.tag == "Messageservers":
            for child in element:
                # noinspection PyProtectedMember
                if type(child) == etree._Comment:
                    continue
                if child.tag != "Messageserver":
                    logger.warning(f"Wrong tag: <{child.tag}> - expected <Messageserver>")
                    continue
                message_servers[child.get("uuid")] = {"mshost": child.get("host"), "msserv": child.get("port")}
    return sysids, message_servers


class Landscape(collections.abc.Mapping):
    """SAP System information from GUI Landscape XML files"""

    def __init__(self):
        self._data = {}
        sysids = {}
        message_servers = {}
        for landscape_file in _get_landscape_files():
            logger.verbose(f"Processing {landscape_file}")
            t_sysids, t_message_servers = _process_landscape_xml(landscape_file)
            sysids.update({i: t_sysids[i] for i in t_sysids})
            message_servers.update(t_message_servers)
        for sysid in sysids:
            if sysids[sysid].get("msid"):
                sysids[sysid].update(message_servers[sysids[sysid].pop("msid")])
                sysids[sysid]["sysid"] = sysid
            logger.trace(f"Landscape Entry {sysid}: {sysids[sysid]}")
            self._data[sysid] = sysids[sysid]

    def __getitem__(self, item):
        return self._data[item.upper()]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rfc']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.8.0,<5.0.0', 'netlink-logging>=0.1.0,<0.2.0', 'pyrfc>=2.5.0,<3.0.0']

setup_kwargs = {
    'name': 'netlink-sap-rfc',
    'version': '0.1.1',
    'description': 'Access SAP via RFC',
    'long_description': "# netlink-sap-rfc\n\nTools for SAP RFC\n\n# Connect\n\nThe default .XML files describing the landscape for SAPGUI are used for access via System ID.\n\n### login_sid\n\nReturns a Connection object.\n\n| Parameter | Description                 | Default        |\n|-----------|-----------------------------|----------------|\n| sysid     | System ID (a.k.a. <SID>)    |                |\n| client    | SAP Client (Mandant)        |                |\n| passwd    | Password                    |                |\n| user      | SAP user ID (BNAME)         | Logged on user |\n| language  | Language                    | EN             |\n| raw       | Don't convert date / time   | False          |\n\n\n### sso\n\nReturns a Connection object.\n\n| Parameter | Description                 | Default        |\n|-----------|-----------------------------|----------------|\n| sysid     | System ID (a.k.a. <SID>)    |                |\n| client    | SAP Client (Mandant)        |                |\n| user      | SAP user ID (BNAME)         | Logged on user |\n| language  | Language                    | EN             |\n| raw       | Don't convert date / time   | False          |\n\n\n # Connection Object\n\nUse one of the functions above to instantiate.\n\n## Methods\n\nAny Remote enabled Function Module can be called as a method (case-insensitive).\n\n### close()\n\nClose the connection. Should always be called before the program is finished, otherwise an error will be logged on the SAP system.\n\n### select(table, *args, **kwargs)\n\nGet contents of a table using `RFC_READ_TABLE`. Returns a list of Records.\n\nColumns within a Record can be access by index (int), name (case-insensitive), or attribute (case-insensitive).\n\n**table** (required) table name (case-insensitive)\n\n***args** (optional) when specified, only the columns listed will be returned\n\n****kwars** (optional) select rows (only equality is supported)\n\n#### Example\n\nrecords = rfc_connection.select('t000', 'mtext', 'ort01', mandt='000')\n\nwould return a list with one item:\n\n```python\n>>> records = c.select('t000', 'mtext', 'ort01', mandt='000')\n>>> len(records)\n1\n>>> records[0][0]\n'SAP SE'\n>>> records[0][1]\n'Walldorf'\n>>> records[0]['mtext']\n'SAP SE'\n>>> records[0]['ort01']\n'Walldorf'\n>>> records[0].mtext\n'SAP SE'\n>>> records[0].ort01\n'Walldorf'\n```\n\n\n\n",
    'author': 'Bernhard Radermacher',
    'author_email': 'bernhard.radermacher@netlink-consulting.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/netlink-consulting/netlink-sap-rfc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
